#include "config.h"

#include <string.h>	// INIT_OBJ
#include <cache/cache.h>

#include "vcc_pesi_debug_if.h"

/*
 * locking: because we require the PRIV_TOP to be created exactly once and
 * varnish-cache takes care of serializing access to the PRIV_TOP tree itself,
 * we do not need to lock here.
 *
 * BUT: For any real-world vmod for which a PRIV_TOP max only be created at esi
 * level > 0, a static mutex is required for access to the priv_top priv (a lock
 * _in_ the priv is not sufficient because also access to the private pointer
 * can race)
 */

struct pedbg_top {
	unsigned		magic;
#define PEDBG_TOP_MAGIC	0x94e33a13
	const struct ws	*ws;
	const struct req	*req;
};

struct pedbg_task {
	unsigned		magic;
#define PEDBG_TASK_MAGIC	0x42a71fb0
	const struct ws	*ws;
	const struct req	*req;
};

#define PEDBG_SETUP(ctx, ptop, ptask, r, tr)	   \
	CHECK_OBJ_NOTNULL(ctx, VRT_CTX_MAGIC);	   \
	AN(ptop);				   \
	AN(ptask);				   \
	r = ctx->req;				   \
	CHECK_OBJ_NOTNULL(r, REQ_MAGIC);	   \
	CHECK_OBJ_NOTNULL(req->top, REQTOP_MAGIC); \
	assert((r)->ws == (ctx)->ws);		   \
	tr = req->top->topreq;			   \
	CHECK_OBJ_NOTNULL(tr, REQ_MAGIC)

/*
 * must be called exactly once
 * restarts check done in vmod for simplicty
 */
VCL_VOID
pedbg_register_privs(VRT_CTX,
    struct vmod_priv *priv_top, struct vmod_priv *priv_task)
{
	struct req *req, *topreq;
	struct pedbg_top *top;
	struct pedbg_task *task;

	PEDBG_SETUP(ctx, priv_top, priv_task, req, topreq);

	if (req->restarts > 0) {
		pedbg_check_privs(ctx, priv_top, priv_task);
		return;
	}

	if (req->esi_level == 0) {
		AZ(priv_top->priv);
		top = WS_Alloc(ctx->ws, sizeof *top);
		AN(top);
		INIT_OBJ(top, PEDBG_TOP_MAGIC);
		top->ws = ctx->ws;
		top->req = req;
		priv_top->priv = top;
	}
	AZ(priv_task->priv);
	task = WS_Alloc(ctx->ws, sizeof *task);
	AN(task);
	INIT_OBJ(task, PEDBG_TASK_MAGIC);
	task->ws = ctx->ws;
	task->req = req;
	priv_task->priv = task;
	pedbg_check_privs(ctx, priv_top, priv_task);
}

#define A(x) assert(x)

VCL_VOID
pedbg_check_privs(VRT_CTX,
    struct vmod_priv *priv_top, struct vmod_priv *priv_task)
{
	struct req *req, *topreq;
	struct pedbg_top *top;
	struct pedbg_task *task;

	PEDBG_SETUP(ctx, priv_top, priv_task, req, topreq);

	CAST_OBJ_NOTNULL(top, priv_top->priv, PEDBG_TOP_MAGIC);
	CAST_OBJ_NOTNULL(task, priv_task->priv, PEDBG_TASK_MAGIC);

	A(task->ws == ctx->ws);
	A(task->req == ctx->req);

	A(top->req == topreq);
	A(top->ws == topreq->ws);

	if (req->esi_level == 0) {
		A(req == topreq);
		A(top->ws == ctx->ws);
		A(top->req == req);
		return;
	}

	A(top->ws != ctx->ws);
	A(top->req != req);
	A(top->req->ws == top->ws);
	A(top->req->ws != ctx->ws);
}
